@@ -18,6 +18,7 @@ import os
 import re
 import sre_constants
 import subprocess
+import warnings
 import io
 from string import Formatter
 from datetime import datetime
@@ -30,7 +31,7 @@ import codecs
 if sys.version_info[0] == 2:
     sys.stdout = codecs.getwriter('utf-8')(sys.stdout)
 
-__VERSION__ = '0.4.1'
+__VERSION__ = '0.5.0'
 
 DESCRIPTION = 'bumpversion: v{} (using Python v{})'.format(
     __VERSION__,
@@ -56,6 +57,11 @@ class DiscardDefaultIfSpecifiedAppendAction(_AppendAction):
         super(DiscardDefaultIfSpecifiedAppendAction, self).__call__(
                 parser, namespace, values, option_string=None)
 
+time_context = {
+    'now': datetime.now(),
+    'utcnow': datetime.utcnow(),
+}
+
 class BaseVCS(object):
 
     @classmethod
@@ -98,8 +104,9 @@ class Git(BaseVCS):
         ]
 
         if lines:
-            assert False, "Git working directory not clean:\n{}".format(
-                b"\n".join(lines))
+            raise WorkingDirectoryIsDirtyException(
+                "Git working directory is not clean:\n{}".format(
+                    b"\n".join(lines)))
 
     @classmethod
     def latest_tag_info(cls):
@@ -165,8 +172,9 @@ class Mercurial(BaseVCS):
         ]
 
         if lines:
-            assert False, "Mercurial working directory not clean:\n{}".format(
-                b"\n".join(lines))
+            raise WorkingDirectoryIsDirtyException(
+                "Mercurial working directory is not clean:\n{}".format(
+                    b"\n".join(lines)))
 
     @classmethod
     def add_path(cls, path):
@@ -182,34 +190,171 @@ VCS = [Git, Mercurial]
 def prefixed_environ():
     return dict((("${}".format(key), value) for key, value in os.environ.items()))
 
+class ConfiguredFile(object):
 
-class VersionPart(object):
+    def __init__(self, path, versionconfig):
+        self.path = path
+        self._versionconfig = versionconfig
+
+    def should_contain_version(self, version, context):
+
+        context['current_version'] = self._versionconfig.serialize(version, context)
+
+        serialized_version = self._versionconfig.search.format(**context)
+
+        if self.contains(serialized_version):
+            return
+
+        msg = "Did not find '{}' or '{}' in file {}".format(version.original, serialized_version, self.path)
+
+        if version.original:
+            assert self.contains(version.original), msg
+            return
+
+        assert False, msg
+
+    def contains(self, search):
+        with io.open(self.path, 'rb') as f:
+            for lineno, line in enumerate(f.readlines()):
+                if search in line.decode('utf-8'):
+                    logger.info("Found '{}' in {} at line {}: {}".format(
+                        search, self.path, lineno, line.decode('utf-8').rstrip()))
+                    return True
+        return False
+
+    def replace(self, current_version, new_version, context, dry_run):
+
+        with io.open(self.path, 'rb') as f:
+            file_content_before = f.read().decode('utf-8')
+
+        context['current_version'] = self._versionconfig.serialize(current_version, context)
+        context['new_version'] = self._versionconfig.serialize(new_version, context)
+
+        search_for = self._versionconfig.search.format(**context)
+        replace_with = self._versionconfig.replace.format(**context)
+
+        file_content_after = file_content_before.replace(
+            search_for, replace_with
+        )
+
+        if file_content_before == file_content_after:
+            # TODO expose this to be configurable
+            file_content_after = file_content_before.replace(
+                current_version.original,
+                replace_with,
+            )
+
+        if file_content_before != file_content_after:
+            logger.info("{} file {}:".format(
+                "Would change" if dry_run else "Changing",
+                self.path,
+            ))
+            logger.info("\n".join(list(unified_diff(
+                file_content_before.splitlines(),
+                file_content_after.splitlines(),
+                lineterm="",
+                fromfile="a/"+self.path,
+                tofile="b/"+self.path
+            ))))
+        else:
+            logger.info("{} file {}".format(
+                "Would not change" if dry_run else "Not changing",
+                self.path,
+            ))
+
+        if not dry_run:
+            with io.open(self.path, 'wt', encoding='utf-8') as f:
+                f.write(file_content_after)
+
+    def __str__(self):
+        return self.path
+
+    def __repr__(self):
+        return '<bumpversion.ConfiguredFile:{}>'.format(self.path)
+
+class VersionPartConfiguration(object):
+    pass
+
+class ConfiguredVersionPartConfiguration(object):
+
+    def __init__(self, values, optional_value=None, first_value=None):
+
+        assert len(values) > 0
+
+        self._values = values
+
+        if optional_value is None:
+            optional_value = values[0]
+
+        assert optional_value in values
+
+        self.optional_value = optional_value
+
+        if first_value is None:
+            first_value = values[0]
+
+        assert first_value in values
+
+        self.first_value = first_value
+
+    def bump(self, value):
+        return self._values[self._values.index(value)+1]
+
+class NumericVersionPartConfiguration(VersionPartConfiguration):
+
+    optional_value = "0"
 
     FIRST_NUMERIC = re.compile('([^\d]*)(\d+)(.*)')
 
-    def __init__(self, value):
+    def __init__(self, first_value=None):
+
+        if first_value is None:
+            first_value = "0"
+
+        int(first_value) # make sure that it's numeric
+
+        self.first_value = first_value
+
+    @classmethod
+    def bump(cls, value):
+        part_prefix, numeric_version, part_suffix = cls.FIRST_NUMERIC.search(value).groups()
+        bumped_numeric = str(int(numeric_version) + 1)
+        return "".join([part_prefix, bumped_numeric, part_suffix])
+
+class VersionPart(object):
+
+    def __init__(self, value, config=None):
         self._value = value
 
+        if config is None:
+            config = NumericVersionPartConfiguration()
+
+        self.config = config
+
     @property
     def value(self):
-        return self._value or "0"
+        return self._value or self.config.optional_value
+
+    def copy(self):
+        return VersionPart(self._value)
 
     def bump(self):
-        part_prefix, numeric_version, part_suffix = self.FIRST_NUMERIC.search(self.value).groups()
-        bumped_numeric = str(int(numeric_version) + 1)
-        self._value = "".join([part_prefix, bumped_numeric, part_suffix])
+        return VersionPart(self.config.bump(self.value), self.config)
 
     def is_optional(self):
-        return self.value == "0"
+        return self.value == self.config.optional_value
 
     def __format__(self, format_spec):
         return self.value
 
     def __repr__(self):
-        return '<bumpversion.VersionPart:{}>'.format(self.value)
+        return '<bumpversion.VersionPart:{}:{}>'.format(
+            self.config.__class__.__name__,
+            self.value
+        )
 
-    def zero(self):
-        self._value = "0"
+    def null(self):
+        return VersionPart(self.config.first_value, self.config)
 
 class IncompleteVersionRepresenationException(Exception):
     def __init__(self, message):
@@ -219,26 +364,73 @@ class MissingValueForSerializationException(Exception):
     def __init__(self, message):
         self.message = message
 
+class WorkingDirectoryIsDirtyException(Exception):
+    def __init__(self, message):
+        self.message = message
+
+def keyvaluestring(d):
+    return ", ".join("{}={}".format(k, v) for k, v in sorted(d.items()))
+
 class Version(object):
 
+    def __init__(self, values, original=None):
+        self._values = dict(values)
+        self.original = original
+
+    def __getitem__(self, key):
+        return self._values[key]
+
+    def __len__(self):
+        return len(self._values)
+
+    def __iter__(self):
+        return iter(self._values)
+
+    def __repr__(self):
+        return '<bumpversion.Version:{}>'.format(keyvaluestring(self._values))
+
+    def bump(self, part_name, order):
+        bumped = False
+
+        new_values = {}
+
+        for label in order:
+            if not label in self._values:
+                continue
+            elif label == part_name:
+                new_values[label] = self._values[label].bump()
+                bumped = True
+            elif bumped:
+                new_values[label] = self._values[label].null()
+            else:
+                new_values[label] = self._values[label].copy()
+
+        new_version = Version(new_values)
+
+        return new_version
+
+class VersionConfig(object):
+
     """
     Holds a complete representation of a version string
     """
 
-    def __init__(self, parse_regex, serialize_formats, context=None):
+    def __init__(self, parse, serialize, search, replace, part_configs=None):
 
         try:
-            self.parse_regex = re.compile(parse_regex)
+            self.parse_regex = re.compile(parse, re.VERBOSE)
         except sre_constants.error as e:
-            logger.error("--parse '{}' is not a valid regex".format(parse_regex))
+            logger.error("--parse '{}' is not a valid regex".format(parse))
             raise e
 
-        self.serialize_formats = serialize_formats
+        self.serialize_formats = serialize
 
-        if not context:
-            context = {}
+        if not part_configs:
+            part_configs = {}
 
-        self.context = context
+        self.part_configs = part_configs
+        self.search = search
+        self.replace = replace
 
     def _labels_for_format(self, serialize_format):
         return (label for _, label, _, _ in Formatter().parse(serialize_format))
@@ -248,37 +440,39 @@ class Version(object):
         # this seems like a good idea because this should be the most complete format
         return self._labels_for_format(self.serialize_formats[0])
 
-    def register_part(self, part):
-        pass
-
     def parse(self, version_string):
 
-        logger.info("Parsing current version '{}' with '{}'".format(version_string, self.parse_regex.pattern))
+        regexp_one_line = "".join([l.split("#")[0].strip() for l in self.parse_regex.pattern.splitlines()])
+
+        logger.info("Parsing version '{}' using regexp '{}'".format(version_string, regexp_one_line))
 
         match = self.parse_regex.search(version_string)
 
-        self._parsed = {}
+        _parsed = {}
         if not match:
             logger.warn("Evaluating 'parse' option: '{}' does not parse current version '{}'".format(
                 self.parse_regex.pattern, version_string))
             return
 
         for key, value in match.groupdict().items():
-            self._parsed[key] = VersionPart(value)
+            _parsed[key] = VersionPart(value, self.part_configs.get(key))
 
-        logger.info("Parsed the following values: " + self._keyvalue_string())
 
-    def _keyvalue_string(self):
-        return ", ".join("{}={}".format(k, v) for k, v in sorted(self._parsed.items()))
+        v = Version(_parsed, version_string)
 
-    def _serialize(self, serialize_format, raise_if_incomplete=False):
+        logger.info("Parsed the following values: %s" % keyvaluestring(v._values))
+
+        return v
+
+    def _serialize(self, version, serialize_format, context, raise_if_incomplete=False):
         """
         Attempts to serialize a version with the given serialization format.
 
         Raises MissingValueForSerializationException if not serializable
         """
-        values = self.context.copy()
-        values.update(self._parsed)
+        values = context.copy()
+        for k in version:
+            values[k] = version[k]
 
         # TODO dump complete context on debug level
 
@@ -293,12 +487,11 @@ class Version(object):
             )
             raise MissingValueForSerializationException(
                 "Did not find key {} in {} when serializing version number".format(
-                    repr(missing_key), repr(self._parsed)))
+                    repr(missing_key), repr(version)))
 
-        keys_needing_representation = set([k for k, v in self._parsed.items() if not v.is_optional()])
-
-        keys_needing_representation = set([])
+        keys_needing_representation = set()
         found_required = False
+
         for k in self.order():
             v = values[k]
 
@@ -317,65 +510,47 @@ class Version(object):
 
         # try whether all parsed keys are represented
         if raise_if_incomplete:
-            if keys_needing_representation > required_by_format:
+            if not (keys_needing_representation <= required_by_format):
                 raise IncompleteVersionRepresenationException(
                     "Could not represent '{}' in format '{}'".format(
-                        "', '".join(keys_needing_representation - required_by_format),
+                        "', '".join(keys_needing_representation ^ required_by_format),
                         serialize_format,
                     ))
 
         return serialized
 
 
-    def _choose_serialize_format(self):
+    def _choose_serialize_format(self, version, context):
 
         chosen = None
 
-        logger.info("Available serialization formats: '{}'".format("', '".join(self.serialize_formats)))
+        # logger.info("Available serialization formats: '{}'".format("', '".join(self.serialize_formats)))
 
         for serialize_format in self.serialize_formats:
             try:
-                self._serialize(serialize_format, raise_if_incomplete=True)
+                self._serialize(version, serialize_format, context, raise_if_incomplete=True)
                 chosen = serialize_format
-                logger.info("Found '{}' to be a usable serialization format".format(chosen))
+                # logger.info("Found '{}' to be a usable serialization format".format(chosen))
             except IncompleteVersionRepresenationException as e:
-                logger.info(e.message)
+                # logger.info(e.message)
                 if not chosen:
                     chosen = serialize_format
             except MissingValueForSerializationException as e:
                 logger.info(e.message)
+                raise e
 
         if not chosen:
             raise KeyError("Did not find suitable serialization format")
 
-        logger.info("Selected serialization format '{}'".format(chosen))
+        # logger.info("Selected serialization format '{}'".format(chosen))
 
         return chosen
 
-    def serialize(self):
-        serialized = self._serialize(self._choose_serialize_format())
-
-        logger.info("Serialized to '{}'".format(serialized))
-
+    def serialize(self, version, context):
+        serialized = self._serialize(version, self._choose_serialize_format(version, context), context)
+        # logger.info("Serialized to '{}'".format(serialized))
         return serialized
 
-    def bump(self, part_name):
-
-        logger.info("Attempting to increment part '{}'".format(part_name))
-
-        bumped = False
-
-        for label in self.order():
-            if not label in self._parsed:
-                continue
-            elif label == part_name:
-                self._parsed[part_name].bump()
-                bumped = True
-            elif bumped:
-                self._parsed[label].zero()
-
-        logger.info("Values are now: " + self._keyvalue_string())
-
 OPTIONAL_ARGUMENTS_THAT_TAKE_VALUES = [
     '--config-file',
     '--current-version',
@@ -383,6 +558,8 @@ OPTIONAL_ARGUMENTS_THAT_TAKE_VALUES = [
     '--new-version',
     '--parse',
     '--serialize',
+    '--search',
+    '--replace',
     '--tag-name',
 ]
 
@@ -414,6 +591,9 @@ def main(original_args=None):
       sys.argv[1:] if original_args is None else original_args
     )
 
+    if len(positionals[1:]) > 2:
+        warnings.warn("Giving multiple files on the command line will be deprecated, please use [bumpversion:file:...] in a config file.", PendingDeprecationWarning)
+
     parser1 = argparse.ArgumentParser(add_help=False)
 
     parser1.add_argument(
@@ -428,6 +608,10 @@ def main(original_args=None):
         '--list', action='store_true', default=False,
         help='List machine readable information', required=False)
 
+    parser1.add_argument(
+        '--allow-dirty', action='store_true', default=False,
+        help="Don't abort if working directory is dirty", required=False)
+
     known_args, remaining_argv = parser1.parse_known_args(args)
 
     logformatter = logging.Formatter('%(message)s')
@@ -470,6 +654,10 @@ def main(original_args=None):
 
     config_file_exists = os.path.exists(known_args.config_file)
 
+    part_configs = {}
+
+    files = []
+
     if config_file_exists:
 
         config.readfp(io.open(known_args.config_file, 'rt', encoding='utf-8'))
@@ -480,6 +668,12 @@ def main(original_args=None):
         logger.info("Reading config file {}:".format(known_args.config_file))
         logger.info(log_config.getvalue())
 
+        if 'files' in dict(config.items("bumpversion")):
+            warnings.warn(
+                "'files =' configuration is will be deprecated, please use [bumpversion:file:...]",
+                PendingDeprecationWarning
+            )
+
         defaults.update(dict(config.items("bumpversion")))
 
         for listvaluename in ("serialize",):
@@ -496,6 +690,50 @@ def main(original_args=None):
             except NoOptionError:
                 pass  # no default value then ;)
 
+        for section_name in config.sections():
+
+            section_name_match = re.compile("^bumpversion:(file|part):(.+)").match(section_name)
+
+            if not section_name_match:
+                continue
+
+            section_prefix, section_value = section_name_match.groups()
+
+            section_config = dict(config.items(section_name))
+
+            if section_prefix == "part":
+
+                ThisVersionPartConfiguration = NumericVersionPartConfiguration
+
+                if 'values' in section_config:
+                    section_config['values'] = list(filter(None, (x.strip() for x in section_config['values'].splitlines())))
+                    ThisVersionPartConfiguration = ConfiguredVersionPartConfiguration
+
+                part_configs[section_value] = ThisVersionPartConfiguration(**section_config)
+
+            elif section_prefix == "file":
+
+                filename = section_value
+
+                if 'serialize' in section_config:
+                    section_config['serialize'] = list(filter(None, (x.strip() for x in section_config['serialize'].splitlines())))
+
+                section_config['part_configs'] = part_configs
+
+                if not 'parse' in section_config:
+                    section_config['parse'] = defaults.get("parse", '(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)')
+
+                if not 'serialize' in section_config:
+                    section_config['serialize'] = defaults.get('serialize', [str('{major}.{minor}.{patch}')])
+
+                if not 'search' in section_config:
+                    section_config['search'] = defaults.get("search", '{current_version}')
+
+                if not 'replace' in section_config:
+                    section_config['replace'] = defaults.get("replace", '{new_version}')
+
+                files.append(ConfiguredFile(filename, VersionConfig(**section_config)))
+
     else:
         message = "Could not read config file at {}".format(known_args.config_file)
         if known_args.config_file != parser1.get_default('config_file'):
@@ -515,6 +753,12 @@ def main(original_args=None):
                          action=DiscardDefaultIfSpecifiedAppendAction,
                          help='How to format what is parsed back to a version',
                          default=defaults.get("serialize", [str('{major}.{minor}.{patch}')]))
+    parser2.add_argument('--search', metavar='SEARCH',
+                         help='Template for complete string to search',
+                         default=defaults.get("search", '{current_version}'))
+    parser2.add_argument('--replace', metavar='REPLACE',
+                         help='Template for complete string to replace',
+                         default=defaults.get("replace", '{new_version}'))
 
     known_args, remaining_argv = parser2.parse_known_args(args)
 
@@ -522,28 +766,30 @@ def main(original_args=None):
 
     assert type(known_args.serialize) == list
 
-    time_context = {
-        'now': datetime.now(),
-        'utcnow': datetime.utcnow(),
-    }
+    context = dict(list(time_context.items()) + list(prefixed_environ().items()) + list(vcs_info.items()))
 
     try:
-        v = Version(
-            known_args.parse,
-            known_args.serialize,
-            context=dict(list(time_context.items()) + list(prefixed_environ().items()) + list(vcs_info.items()))
+        vc = VersionConfig(
+            parse=known_args.parse,
+            serialize=known_args.serialize,
+            search=known_args.search,
+            replace=known_args.replace,
+            part_configs=part_configs,
         )
     except sre_constants.error as e:
         sys.exit(1)
 
+    current_version = vc.parse(known_args.current_version) if known_args.current_version else None
+
+    new_version = None
+
     if not 'new_version' in defaults and known_args.current_version:
-        v.parse(known_args.current_version)
-
-        if len(positionals) > 0:
-            v.bump(positionals[0])
-
         try:
-            defaults['new_version'] = v.serialize()
+            if current_version and len(positionals) > 0:
+                logger.info("Attempting to increment part '{}'".format(positionals[0]))
+                new_version = current_version.bump(positionals[0], vc.order())
+                logger.info("Values are now: " + keyvaluestring(new_version._values))
+                defaults['new_version'] = vc.serialize(new_version, context)
         except MissingValueForSerializationException as e:
             logger.info("Opportunistic finding of new_version failed: " + e.message)
         except IncompleteVersionRepresenationException as e:
@@ -591,75 +837,58 @@ def main(original_args=None):
                          help='Commit message',
                          default=defaults.get('message', 'Bump version: {current_version} → {new_version}'))
 
-    files = []
+
+    file_names = []
     if 'files' in defaults:
         assert defaults['files'] != None
-        files = defaults['files'].split(' ')
+        file_names = defaults['files'].split(' ')
 
     parser3.add_argument('part',
                          help='Part of the version to be bumped.')
     parser3.add_argument('files', metavar='file',
                          nargs='*',
-                         help='Files to change', default=files)
+                         help='Files to change', default=file_names)
 
     args = parser3.parse_args(remaining_argv + positionals)
 
     if args.dry_run:
         logger.info("Dry run active, won't touch any files.")
 
+    if not new_version:
+        new_version = vc.parse(args.new_version)
+
     logger.info("New version will be '{}'".format(args.new_version))
 
-    files = files or positionals[1:]
+    file_names = file_names or positionals[1:]
 
+    for file_name in file_names:
+        files.append(ConfiguredFile(file_name, vc))
 
     for vcs in VCS:
         if vcs.is_usable():
-            vcs.assert_nondirty()
+            try:
+                vcs.assert_nondirty()
+            except WorkingDirectoryIsDirtyException as e:
+                if not defaults['allow_dirty']:
+                    logger.warn(
+                        "{}\n\nUse --allow-dirty to override this if you know what you're doing.".format(e.message))
+                    raise
             break
         else:
             vcs = None
 
     # make sure files exist and contain version string
 
-    logger.info("Asserting files {} contain string '{}':".format(", ".join(files), args.current_version))
+    logger.info("Asserting files {} contain the version string:".format(", ".join([str(f) for f in files])))
 
-    for path in files:
-        with io.open(path, 'rb') as f:
-
-            found_before = False
-
-            for lineno, line in enumerate(f.readlines()):
-                if args.current_version in line.decode('utf-8'):
-                    found_before = True
-                    logger.info("Found '{}' in {} at line {}: {}".format(args.current_version, path, lineno, line.decode('utf-8').rstrip()))
-
-            assert found_before, 'Did not find string {} in file {}'.format(
-                args.current_version, path)
+    for f in files:
+        f.should_contain_version(current_version, context)
 
     # change version string in files
-    for path in files:
-        with io.open(path, 'rb') as f:
-            before = f.read().decode('utf-8')
+    for f in files:
+        f.replace(current_version, new_version, context, args.dry_run)
 
-        after = before.replace(args.current_version, args.new_version)
-
-        logger.info("{} file {}:".format(
-            "Would change" if args.dry_run else "Changing",
-            path,
-        ))
-        logger.info("\n".join(list(unified_diff(
-            before.splitlines(),
-            after.splitlines(),
-            lineterm="",
-            fromfile="a/"+path,
-            tofile="b/"+path
-        ))))
-
-        if not args.dry_run:
-            with io.open(path, 'wt', encoding='utf-8') as f:
-                f.write(after)
-
-    commit_files = files
+    commit_files = [f.path for f in files]
 
     config.set('bumpversion', 'new_version', args.new_version)
 
